/*
 * Copyright (C) 1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <stdlib.h>	/* for getenv(3) */
#ifdef EXC_BY_SIGNAL
#  if defined(linux)
#    include <asm/segment.h>
#  endif
#endif

#include "monitor.h"	/* for macro BINCLASS_{,UN}LOCK() */
#include "compiler.h"


/*
 * Global Variables
 */
bool_t compiler_enabled = TRUE;

#if defined(EXC_BY_SIGNAL) && defined(SEARCH_SIGCONTEXT)
#  ifdef linux
unsigned short reg_gs;
unsigned short reg_fs;
#  endif
unsigned short reg_es;
unsigned short reg_ds;
#  ifdef __FreeBSD__
unsigned short reg_cs;
unsigned short reg_ss;
#  endif
#endif

int opt_quiet = 0;
int opt_outcode = 0;
int opt_codesize = 0;
int opt_dontcmplvmcls = 0;
int opt_igndisable = 0;
int opt_cmplclinit = 0;

void *sym_compileAndInvokeMethod;
void *sym_invokeJITCompiledMethod;
void *sym_invokeJavaMethod;
void *sym_invokeSynchronizedJavaMethod;
void *sym_invokeAbstractMethod;
void *sym_invokeLazyNativeMethod;


/*
 * Local Functions
 */
static void initializeClass(ClassClass *);
#ifdef __FreeBSD__
static void signalHandler(int sig, int code, void *uc);
#else
static void signalHandler(int sig, void *info, void *uc);
#endif
static void freeClass(ClassClass *);
static unsigned char *compiledCodePC(JavaFrame *, struct methodblock *);
static bool_t pcInCompiledCode(unsigned char *pc, struct methodblock *);
static JavaFrame *framePrev(JavaFrame *frame, JavaFrame *frame_buf);
static void compilerEnable();
static void compilerDisable();
static bool_t compileClass(ClassClass *);
static bool_t compileClasses(Hjava_lang_String *);
static HObject *compilerCommand(HObject *);

#define INITIALIZE_METHOD(MB) \
  if (opt_cmplclinit || strcmp((MB)->fb.name, "<clinit>")) {\
    (MB)->fb.access |= ACC_MACHINE_COMPILED;\
    (MB)->invoker = sym_compileAndInvokeMethod;\
  }


/*
 * Initialization of the compiler.
 * This method is called by JVM.
 */
void java_lang_Compiler_start(void ** vector) {
  static int already_called = 0;
  int version;
  int javaVersion, javaMinorVersion, compilerVersion;

#ifdef COMPILE_DEBUG
  printf("java_lang_Compiler_start() called.\n");
  fflush(stdout);
#endif
  if (already_called)  return;
  already_called = 1;

  /* get options */
  {
    char *opts = getenv("JAVA_COMPILER_OPT");
    char *opt;

    if (opts) {
#ifdef RUNTIME_DEBUG
      printf("  JAVA_COMPILER_OPT: %s\n", opts);
      fflush(stdout);
#endif

      opt = strtok(opts, ", ");
      while (opt) {
	if (!strncmp(opt, "quiet", 6)) {
	  opt_quiet = 1;
	}
	else if (!strncmp(opt, "outcode", 8)) {
	  opt_outcode = 1;
	  if (!opt_quiet)  printf("  option: outcode\n");
	}
	else if (!strncmp(opt, "codesize", 9)) {
	  opt_codesize = 1;
	  if (!opt_quiet)  printf("  option: codesize\n");
	}
	else if (!strncmp(opt, "dontcmplvmcls", 14)) {
	  opt_dontcmplvmcls = 1;
	  if (!opt_quiet)  printf("  option: dontcmplvmcls\n");
	}
	else if (!strncmp(opt, "igndisable", 11)) {
	  opt_igndisable = 1;
	  if (!opt_quiet)  printf("  option: igndisable\n");
	}
	else if (!strncmp(opt, "cmplclinit", 11)) {
	  opt_cmplclinit = 1;
	  if (!opt_quiet)  printf("  option: cmplclinit\n");
	}
	opt = strtok(NULL, ", ");
      }
    }
  }


  /* show credit */
  if (!opt_quiet) {
    fprintf(stderr, CREDIT);
    fflush(stderr);
  }


  /* version check */

  version = *(int *)vector[0];
  compilerVersion = version >> 24;
  javaVersion = (version >> 16) & 0xff;
  javaMinorVersion = version & 0xffff;

  if (compilerVersion != COMPILER_VERSION) {
    printf("warning: version num. of compiler interface is not %d: %d\n",
		COMPILER_VERSION, compilerVersion);
  }
  if ( (javaVersion != JAVA_VERSION) ||
       (javaMinorVersion != JAVA_MINOR_VERSION) ) {
    printf("warning: version num. of class file format is not %d.%d: %d.%d\n",
	JAVA_VERSION, JAVA_MINOR_VERSION, javaVersion, javaMinorVersion);
  }


  /* initialization */

#if defined(EXC_BY_SIGNAL) && defined(SEARCH_SIGCONTEXT)
#ifdef linux
  __asm__("movw %%gs,%0" : "=m" (reg_gs));
  __asm__("movw %%fs,%0" : "=m" (reg_fs));
#endif
  __asm__("movw %%es,%0" : "=m" (reg_es));
  __asm__("movw %%ds,%0" : "=m" (reg_ds));
#ifdef __FreeBSD__
  __asm__("movw %%cs,%0" : "=m" (reg_cs));
  __asm__("movw %%ss,%0" : "=m" (reg_ss));
#endif

#ifdef RUNTIME_DEBUG
#ifdef linux
  printf("segment registers (gs,fs,es,ds): %04x, %04x, %04x, %04x\n",
	reg_gs, reg_fs, reg_es, reg_ds);
#endif	/* linux */
#ifdef __FreeBSD__
  printf("segment registers (es,ds,cs,ss): %04x, %04x, %04x, %04x\n",
	reg_es, reg_ds, reg_cs, reg_ss);
#endif	/* __FreeBSD__ */
  fflush(stdout);
#endif
#endif	/* EXC_BY_SIGNAL && SEARCH_SIGCONTEXT */

#ifdef __FreeBSD__
  /* to get addresses of functions in lib[cm] with sysDynamicLink() */
  /* libc for __divdi3, __moddi3, printf, fflush */
  if (!sysAddDLSegment("/usr/lib/libc.so.3.1")) {
    if (!sysAddDLSegment("/usr/lib/libc.so.3.0")) {
      printf("fatal: JIT compiler can't find libc.so.3.1 or 3.0.\n"
		"  You must edit compiler.c and re-compile"
		"  if you use libc on your system.\n");
      sysExit(1);
    }
  }
  /* libm for fmod */
  if (!sysAddDLSegment("/usr/lib/libm.so.2.0")) {
    printf("fatal: JIT compiler can't find libm.so.2.0.\n"
		"  You must edit compiler.c and re-compile"
		"  if you use libc on your system.\n");
    sysExit(1);
  }
#endif

  /* resolve some symbols */
  sym_compileAndInvokeMethod =
		(void *)sysDynamicLink("compileAndInvokeMethod");
  sym_invokeJITCompiledMethod =
		(void *)sysDynamicLink("invokeJITCompiledMethod");
  sym_invokeJavaMethod = (void *)sysDynamicLink("invokeJavaMethod");
  sym_invokeSynchronizedJavaMethod =
		(void *)sysDynamicLink("invokeSynchronizedJavaMethod");
  sym_invokeAbstractMethod = (void *)sysDynamicLink("invokeAbstractMethod");
  sym_invokeLazyNativeMethod =
		(void *)sysDynamicLink("invokeLazyNativeMethod");
#ifdef COMPILE_DEBUG
  printf("symbols:\n");
  printf("  compileAndInvokeMethod: 0x%08x\n",
		(int)sym_compileAndInvokeMethod);
  printf("  invokeJITCompiledMethod: 0x%08x\n",
		(int)sym_invokeJITCompiledMethod);
  printf("  invokeJavaMethod: 0x%08x\n", (int)sym_invokeJavaMethod);
  printf("  invokeSynchronizedJavaMethod: 0x%08x\n",
		(int)sym_invokeSynchronizedJavaMethod);
  printf("  invokeAbstractMethod: 0x%08x\n", (int)sym_invokeAbstractMethod);
  printf("  invokeLazyNativeMethod: 0x%08x\n",
		(int)sym_invokeLazyNativeMethod);
#endif

#ifdef RESOLVE_SYMBOL_ON_CODE
  /* resolve function symbols in pre-assembled code (assembledCode()) */
  initFunctionSymbols();
#endif

  /* compile the classes which is already loaded */
  if (!opt_dontcmplvmcls) {
    ClassClass **clazzptr;
    int i;

    BINCLASS_LOCK();
    for (i = nbinclasses, clazzptr = binclasses; --i >= 0; clazzptr++) {
#ifdef COMPILE_DEBUG
      printf("classes already in VM: %s\n", cbName(*clazzptr));
      fflush(stdout);
#endif
      initializeClass(*clazzptr);
    }
    BINCLASS_UNLOCK();
  }

  /* prohibit generating lossy opcodes */
  UseLosslessQuickOpcodes = TRUE;

  /* set up link vector */
  *(char **)vector[1] = (char *)initializeClass;
  *(char **)vector[2] = (char *)sym_invokeJITCompiledMethod;
  *(char **)vector[3] = (char *)signalHandler;
  *(char **)vector[4] = (char *)freeClass;
  *(char **)vector[5] = (char *)compileClass;
  *(char **)vector[6] = (char *)compileClasses;
  if (!opt_igndisable) {
    *(char **)vector[7] = (char *)compilerEnable;
    *(char **)vector[8] = (char *)compilerDisable;
  }
  *(char **)vector[10] = (char *)pcInCompiledCode;
  *(char **)vector[11] = (char *)compiledCodePC;
  *(char **)vector[70] = (char *)framePrev;

#ifdef COMPILE_DEBUG
  {
    int i;
    for (i = 1; i <= 8; i++)
      printf("*vector[%d]: 0x%08x\n", i, (int)*(char **)vector[i]);
    printf("*vector[10]: 0x%08x\n", i, (int)*(char **)vector[10]);
    printf("*vector[11]: 0x%08x\n", i, (int)*(char **)vector[11]);
    printf("*vector[70]: 0x%08x\n", i, (int)*(char **)vector[70]);
    fflush(stdout);
  }
#endif

#ifdef COMPILE_DEBUG
  printf("java_lang_Compiler_start() done.\n");
  fflush(stdout);
#endif
}


/*
 * Local Functions
 */

/*
 * Initialize the class at the time one is loaded.
 */
void initializeClass(ClassClass *cb) {
  struct methodblock *mb, *mb_end;
#ifdef COMPILE_DEBUG
  printf("initializeClass(%s) called.\n", cbName(cb));
  fflush(stdout);
#endif

  mb = cbMethods(cb);
  mb_end = mb + cbMethodsCount(cb);
  for (; mb < mb_end; mb++) {
    /* initialize mb->CompiledCodeInfo */
    if (!(mb->CompiledCodeInfo))  prepareCompiledCodeInfo(EE(), mb);

    if (mb->fb.access & (ACC_ABSTRACT | ACC_NATIVE))
	/* methods not to be compiled */
	continue;

    INITIALIZE_METHOD(mb);
  }
}


#ifdef __FreeBSD__
static void signalHandler(int sig, int code, void *uc) {
#else
static void signalHandler(int sig, void *info, void *uc) {
#endif
#ifdef EXC_BY_SIGNAL
#  ifdef SEARCH_SIGCONTEXT
  int32_t first_local_variable;
  static int sc_offset = 0;
  int32_t *ptr = ((int32_t *)&first_local_variable) + 4;
	/* 4: first...variable, esp, 2 arguments */
#  endif	/* SEARCH_SIGCONTEXT */
  sigcontext *sc;
  struct methodblock *mb;
  CodeInfo *codeinfo;
  int native_off;
  throwentry *tentry;
#endif	/* EXC_BY_SIGNAL */

#ifdef RUNTIME_DEBUG
  printf("Signal handler for compiled code called. (sig: %d)\n", sig);
  showStackFrames(EE());
#endif

#ifdef EXC_BY_SIGNAL
#ifdef SEARCH_SIGCONTEXT
  /* get signal context */
  ptr = ((int32_t *)&first_local_variable) + 4;
	/* 4: first...variable, esp, 2 arguments */
  if (sc_offset > 0) {
    sc = (sigcontext *)(ptr + sc_offset);
  }
  else {	/* must search */
#define SEARCH_WIDTH 50
    for (sc_offset = 0; sc_offset < SEARCH_WIDTH - 3; sc_offset++) {
#ifdef __FreeBSD__
      if ( (*((unsigned short *)(ptr)) == reg_es) &&
	(*((unsigned short *)(ptr + 1)) == reg_ds) &&
	(*((unsigned short *)(ptr + 2)) == reg_cs) &&
	(*((unsigned short *)(ptr + 3)) == reg_ss) ) {
	sc_offset -= 7;
	ptr -= 7;
	break;
      }
#else
      if ( (*((unsigned short *)(ptr)) == reg_gs) &&
	(*((unsigned short *)(ptr + 1)) == reg_fs) &&
	(*((unsigned short *)(ptr + 2)) == reg_es) &&
	(*((unsigned short *)(ptr + 3)) == reg_ds) )
	break;
#endif
      else
	ptr += 1;
    }
    if (sc_offset >= SEARCH_WIDTH - 3) {
      printf("FATAL: struct sigcontext is not found.");
      exit(1);
    }
    sc = (sigcontext *)ptr;
#ifdef RUNTIME_DEBUG
    printf("sc_offset: %d\n", sc_offset);
#endif
  }
#else	/* SEARCH_SIGCONTEXT*/
  sc = (sigcontext *)uc;
#endif	/* SEARCH_SIGCONTEXT*/
#ifdef RUNTIME_DEBUG
  printf("sc: 0x%08x\n");  fflush(stdout);
  showSigcontext(sc);
#endif


  if ((sig == SIGSEGV) || (sig == SIGBUS)) {
    /* NullPointerException occurred */
    SignalError(EE(), JAVAPKG "NullPointerException", 0);
	/* set ee->exception.exc, exception object */

    mb = (struct methodblock *) *(((int32_t *)sc->SC_EBP) + 3);
	/* mb is second argument of assembledCode(): 0xc(%ebp) */
#ifdef RUNTIME_DEBUG
    printf("method(0x%x): %s#%s %s\n", (int)mb,
	cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature);
    fflush(stdout);
#endif
    codeinfo = (CodeInfo *)(mb->CompiledCodeInfo);
    native_off =
	((unsigned char *)sc->SC_EIP) - ((unsigned char *)mb->CompiledCode);
#ifdef RUNTIME_DEBUG
    printf("native off: 0x%08x\n", native_off);
    fflush(stdout);
#endif
    tentry = throwtableGet(codeinfo, native_off);
#ifdef RUNTIME_DEBUG
    printf("throwentry: ");
    if (tentry)
      printf("start: 0x%x, end: 0x%x, byteoff: 0x%x\n",
		tentry->start, tentry->end, tentry->byteoff);
    else
      printf(" (null)\n");
    fflush(stdout);
#endif

#ifdef __FreeBSD__
    /* re-initialize signal handler */
    intrInitMD();
#endif

    /* set (int32_t)bytepcoff (local variable of assembledCode) */
    *(((int32_t *)sc->SC_EBP) - 1) = tentry->byteoff;

    /* jump to exception handler */
    {
      register uint32_t orig_sp asm("ecx");
      register uint32_t orig_bp asm("edx");
      register uint32_t exc_handler asm("eax");

      exc_handler =
	(uint32_t)(mb->CompiledCode + codeinfo->exc_handler_nativeoff);
      orig_sp = sc->SC_ESP;
      orig_bp = sc->SC_EBP;

      __asm__("movl  %%ecx,%%esp\n\t"
	      "movl  %%edx,%%ebp\n\t"
	      "jmp   *%%eax"
		: : "r" (orig_sp), "r" (orig_bp), "r" (exc_handler));
    }	/* jump to exception handler */
  }	/* if ((sig == SIGSEGV) || (sig == SIGBUS)) */
#endif	/* EXC_BY_SIGNAL */
}


/*
 * Freeing class staffs related to the compiler.
 */
static void freeClass(ClassClass *cb) {
  struct methodblock *mb = cbMethods(cb);
  struct methodblock *mb_end = mb + cbMethodsCount(cb);

#ifdef COMPILE_DEBUG
  printf("freeClass() called.\n");
  fflush(stdout);
#endif

  for (; mb < mb_end; mb++)
    freeMethod(mb);
}


static unsigned char *compiledCodePC(JavaFrame *frame,struct methodblock *mb) {
  return (frame ? frame->lastpc : mb->code);
}


static bool_t pcInCompiledCode(unsigned char *pc, struct methodblock *mb) {
  unsigned int off = pc - mb->code;
  return (off < mb->code_length);
}


static JavaFrame *framePrev(JavaFrame *frame, JavaFrame *frame_buf) {
  return frame->prev;
}


/*
 * Enable and Disable the compiler.
 */
static void compilerEnable() {
#ifdef RUNTIME_DEBUG
  printf("compilerEnable() called.\n");  fflush(stdout);
#endif
  compiler_enabled = TRUE;
}

/*
 * Enable and Disable the compiler.
 */
static void compilerDisable() {
#ifdef RUNTIME_DEBUG
  printf("compilerDisable() called.\n");  fflush(stdout);
#endif
  compiler_enabled = FALSE;
}


/*
 * Compile the class. Not only initialization.
 */
static bool_t compileClass(ClassClass *cb) {
  struct methodblock *mb, *mb_end;

#ifdef COMPILE_DEBUG
  printf("\n");
  printf("compileClass(%s) called.\n", cbName(cb));
  fflush(stdout);
#endif

  if (!compiler_enabled) {
#ifdef COMPILE_DEBUG
    printf("  compiler has been disabled.\n  return.\n");
    fflush(stdout);
#endif
    return TRUE;
  }

  mb = cbMethods(cb);
  mb_end = mb + cbMethodsCount(cb);
  for (; mb < mb_end; mb++) {
    /* initialize mb->CompiledCodeInfo */
    if (!(mb->CompiledCodeInfo))  prepareCompiledCodeInfo(EE(), mb);

    if (!(mb->code)) {
#ifdef COMPILE_DEBUG
      printf("code is empty: %s %s\n", mb->fb.name, mb->fb.signature);
      fflush(stdout);
#endif
      INITIALIZE_METHOD(mb);
      continue;
    }

    if (compileMethod(mb))  return FALSE;
  }

#ifdef COMPILE_DEBUG
  printf("compileClass(%s) done.\n", cbName(cb));
  fflush(stdout);
#endif

  return TRUE;
}


/*
 * Compile the classes specified by comma-delimited names.
 */
static bool_t compileClasses(Hjava_lang_String *nameStr) {
  char *names = allocCString(nameStr);
  char *name = names;
  char *p = names;
  bool_t done = FALSE;

#ifdef COMPILE_DEBUG
  printf("compileClasses() called.\n");
  fflush(stdout);
#endif

  if (names) {
    ClassClass *cb;

    do {
      while ((*p != ',') && (*p != '\0'))  p++;
      if (*p == '\0')  done = TRUE;
      else  *p = '\0';

#ifdef COMPILE_DEBUG
      printf("compileClasses(): classname is `%s'\n", name);
      fflush(stdout);
#endif
      cb = FindClass(NULL, name, TRUE);
      if (cb)
	if (!(compileClass(cb)))  goto compileDone;

      p++;
      name = p;
    } while (!done);

  compileDone:
    sysFree(names);

    return TRUE;
  }

  return FALSE;
}


/*
 * Process the command from Java code.
 */
static HObject *compilerCommand(HObject *obj) {
  HObject *ret = NULL;
  
#ifdef COMPILE_DEBUG
  printf("compilerCommand() called.\n");
  fflush(stdout);
#endif

  /* do nothing. */

  return ret;
}
